"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.runTransaction = exports.treeApi = exports.rollback = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const index_js_1 = require("../feature-libraries/index.js");
const index_js_2 = require("../simple-tree/index.js");
const index_js_3 = require("../util/index.js");
const schematizingTreeView_js_1 = require("./schematizingTreeView.js");
const treeView_js_1 = require("./treeView.js");
/**
 * A special object that signifies when a SharedTree {@link RunTransaction | transaction} should "roll back".
 * @public
 */
exports.rollback = Symbol("SharedTree Transaction Rollback");
/**
 * The `Tree` object holds various functions for interacting with {@link TreeNode}s.
 * @public
 */
exports.treeApi = {
    ...index_js_2.treeNodeApi,
    runTransaction: createRunTransaction(),
    contains(parent, child) {
        let toCheck = child;
        while (toCheck !== undefined) {
            if (toCheck === parent) {
                return true;
            }
            toCheck = exports.treeApi.parent(toCheck);
        }
        return false;
    },
};
// TODO: Add more constraint types here
/** Creates a copy of `runTransaction` with the `rollback` property added so as to satisfy the `RunTransaction` interface. */
function createRunTransaction() {
    /** A type-safe helper to add a "rollback" property (as required by the `RunTransaction` interface) to a given object */
    function defineRollbackProperty(target) {
        Reflect.defineProperty(target, "rollback", { value: exports.rollback });
        return target;
    }
    return defineRollbackProperty(runTransaction.bind({}));
}
/**
 * Run the given transaction.
 * @remarks
 * This API is not publicly exported but is exported outside of this module so that test code may unit test the `Tree.runTransaction` function directly without being restricted to its public API overloads.
 */
function runTransaction(treeOrNode, transaction, preconditions = []) {
    if (treeOrNode instanceof schematizingTreeView_js_1.SchematizingSimpleTreeView) {
        const t = transaction;
        return runTransactionInCheckout(treeOrNode.checkout, () => t(treeOrNode.root), preconditions);
    }
    else {
        const node = treeOrNode;
        const t = transaction;
        const context = (0, index_js_2.getFlexNode)(node).context;
        (0, internal_1.assert)(context instanceof index_js_1.Context, 0x901 /* Unsupported context */);
        const treeView = treeView_js_1.contextToTreeView.get(context) ?? (0, index_js_3.fail)("Expected view to be registered for context");
        return runTransactionInCheckout(treeView.checkout, () => t(node), preconditions);
    }
}
exports.runTransaction = runTransaction;
function runTransactionInCheckout(checkout, transaction, preconditions) {
    checkout.transaction.start();
    for (const constraint of preconditions) {
        switch (constraint.type) {
            case "nodeInDocument": {
                const node = (0, index_js_2.getFlexNode)(constraint.node);
                (0, internal_1.assert)(node.treeStatus() === index_js_1.TreeStatus.InDocument, 0x90f /* Attempted to apply "nodeExists" constraint when building a transaction, but the node is not in the document. */);
                checkout.editor.addNodeExistsConstraint(node.anchorNode);
                break;
            }
            default:
                (0, internal_1.unreachableCase)(constraint.type);
        }
    }
    let result;
    try {
        result = transaction();
    }
    catch (e) {
        // If the transaction has an unhandled error, abort and rollback the transaction but continue to propagate the error.
        checkout.transaction.abort();
        throw e;
    }
    if (result === exports.rollback) {
        checkout.transaction.abort();
    }
    else {
        checkout.transaction.commit();
    }
    return result;
}
//# sourceMappingURL=treeApi.js.map